/* (c) Copyright 1985 Cary Buchman and Daniel M. Berry */
/* (c) Copyright 1987 Berry Computer Scientists, Ltd. */
/************************************************************************* 
**************************************************************************
**                                                                      **
** this file includes routines that dump the current internal line      **
**                                                                      **
************************************************************************** 
*************************************************************************/ 

#include <stdio.h>
#include "TOKEN.h"
#include "lex.h"
#include "macros.h"

#define  max(a,b)       ((a<b) ? b : a)
#define  MAXZWC         100        /* maximum number pf respective zero
                                      width characters        */
#define  MAXNOFONTS          255        /* defined also in width.c */
#define  MAXSTRETCHINGPLACES 100
TOKENPTR new_token();              /* function return type    */

extern   bool stretchable_fonts[256];
extern   char *width_table[MAXNOFONTS+1];
extern   char stretch_mode;
extern   char stretch_places;
extern   double stretch_limit;
extern   int favor_distribution;
extern   int favor_stretch;
extern   int unit_width;
extern   int no_stretchable_fonts;
extern   TOKENPTR stretching_places[MAXSTRETCHINGPLACES];
extern   char stretching_style[MAXSTRETCHINGPLACES];
extern   int connection_stretching_place_index[MAXSTRETCHINGPLACES];
extern   int letter_stretching_place_index[MAXSTRETCHINGPLACES];
extern   int no_stretching_places;
extern   int no_connection_stretching_places;
extern   int no_letter_stretching_places;
extern   bool left_to_right_sweep;
extern   int no_of_words;
extern	 bool no_add_to_manually_stretched_letter;
extern	 bool no_add_to_manually_stretched_connection;
extern	 bool no_add_to_manually_stretched_word;
extern	 bool any_diacriticals;

/************************************************************************* 
************************************************************************** 
**                                                                      **
** this routine dumps the current internal line while reversing         ** 
** the tokens of the specified lr direction                             **
** Was enhanced in order to deal with ZERO width characters, or         **
** ZERO horizontal movements (via the usage of '\z' for example).       **
** It was also optimized and made more efficient.                       **
**                                                                      **
** Johny                                                                **
************************************************************************** 
*************************************************************************/ 

dump_line(start,end,reverse_lr)

  TOKENPTR *start;
  TOKENPTR *end;
  bool     reverse_lr;

{

  bool     start_of_line = TRUE;

  TOKENPTR tmptr;
  TOKENPTR tokenptr  = *start;    /* next token to dump                  */
  TOKENPTR continue_from = NULL;  /* to point to last token before going */
                                  /* into reversing loop to allow        */
                                  /* repairing linked list of tokens     */
  TOKENPTR zerowc[MAXZWC];
  int      total_stretch = 0;
  int      i,j,hpos;
  bool      begining;
  bool      ending;
  int       next_space_width;
  int       prev_space_width;
  int       next_space_spread;
  int       prev_space_spread;
  bool      at_least_one_char_in_stretchable_font=FALSE;

/* fprintf(stderr, "dump_line: %d %d %d\n", start, end, reverse_lr); */
/* printf("dump_line: %d %d %d\n", start, end, reverse_lr);
   printf("at beginning:\n");
   print_line(start,end); */

  no_of_words = 0;

  if (tokenptr != NULL) {
      if (tokenptr->ending)
         no_of_words++;
      if (stretchable_fonts[tokenptr->font])
         at_least_one_char_in_stretchable_font = TRUE;
  }
  while(tokenptr != NULL)
  {
    if  (tokenptr->lr != reverse_lr) 
    /* if the current character's direction == current document direction */
    {
      continue_from = tokenptr;
      tokenptr = tokenptr->next;
      if (tokenptr != NULL) {
          if (tokenptr->ending)
             no_of_words++;
          if (stretchable_fonts[tokenptr->font])
             at_least_one_char_in_stretchable_font = TRUE;
      }
    }
    else
    {
      TOKENPTR rev_start = NULL;     /* ptr to start of reversed token line */
      TOKENPTR rev_end = NULL;       /* ptr to end of reversed token line   */
      bool original_prev_of_future_rev_end_is_a_begining;
      bool future_rev_end_is_now_a_begining;
      bool original_next_of_future_rev_start_is_now_an_ending;
      bool future_rev_start_is_now_an_ending;
      int original_prev_space_width_of_future_rev_end;
      int original_prev_space_spread_of_future_rev_end;
      int original_next_space_width_of_future_rev_start;
      int original_next_space_spread_of_future_rev_start;

      if (continue_from == NULL){
         original_prev_of_future_rev_end_is_a_begining = FALSE;
      } else {
         original_prev_of_future_rev_end_is_a_begining =
	    continue_from->begining;
      }
      future_rev_end_is_now_a_begining = tokenptr->begining;
      original_prev_space_width_of_future_rev_end =
          tokenptr->prev_space_width;
      original_prev_space_spread_of_future_rev_end =
          tokenptr->prev_space_spread;
      
      while ((tokenptr != NULL) && (tokenptr->lr == reverse_lr))
      /* go out when tokenptr points past the end of line or to a token
      whose direction = the document direction */
      {
        /* swap word endings and beginnings */ 
        begining = tokenptr->begining;
        ending   = tokenptr->ending;
        next_space_width = tokenptr->next_space_width;
        prev_space_width = tokenptr->prev_space_width;
        next_space_spread = tokenptr->next_space_spread;
        prev_space_spread = tokenptr->prev_space_spread;
        tokenptr->begining = ending;
        tokenptr->ending   = begining;
        tokenptr->next_space_width = prev_space_width;
        tokenptr->prev_space_width = next_space_width;
        tokenptr->next_space_spread = prev_space_spread;
        tokenptr->prev_space_spread = next_space_spread;
        hpos = tokenptr->horizontal_pos;
        i = 0;
        while ((tokenptr->next != NULL) &&
               (tokenptr->next->horizontal_pos == hpos))
        {
            zerowc[i++] = tokenptr;
            tokenptr = tokenptr->next;
	    if (tokenptr != NULL) {
                if (tokenptr->ending)
                   no_of_words++;
                if (stretchable_fonts[tokenptr->font])
                   at_least_one_char_in_stretchable_font = TRUE;
	    }
        }
        tmptr = tokenptr->next;
        if (i>0)        /* a ZERO-width character was detected */
        {
          int new_first_zw_begining;
          int new_first_zw_prev_space_width;
          int new_first_zw_prev_space_spread;

	  any_diacriticals = TRUE;
          zerowc[i] = tokenptr;
          for (j=i; j>=0; j--) {
            push_token(zerowc[j],&rev_start,&rev_end);
          }
          new_first_zw_begining = zerowc[i]->ending;
          new_first_zw_prev_space_width = zerowc[i]->next_space_width;
          new_first_zw_prev_space_spread = zerowc[i]->next_space_spread;

          zerowc[i]->ending = zerowc[0]->ending;
          zerowc[i]->next_space_width = zerowc[0]->next_space_width;
          zerowc[i]->next_space_spread = zerowc[0]->next_space_spread;

          zerowc[0]->ending = NOT_END;
          zerowc[0]->next_space_width = 0;
          zerowc[0]->next_space_spread = 0;

          zerowc[0]->begining = new_first_zw_begining;
          zerowc[0]->prev_space_width = new_first_zw_prev_space_width;
          zerowc[0]->prev_space_spread = new_first_zw_prev_space_spread;
        }
        else push_token(tokenptr,&rev_start,&rev_end);

        /* doing now each time through the loop in case we exit at the next
        time at the top */
	if (tmptr == NULL){
           original_next_of_future_rev_start_is_now_an_ending = FALSE;
	} else {
           original_next_of_future_rev_start_is_now_an_ending = tmptr->ending;
	}
        /* on the assumption that x->begining and x->ending AND
        x->next_space_spread and x->prev_space_spread,
        x->next_space_width and x->prev_space_width, resp., for x in rev_end
        .. rev_start, have already been swapped */
        future_rev_start_is_now_an_ending = tokenptr->begining;
        original_next_space_width_of_future_rev_start =
           tokenptr->prev_space_width;
        original_next_space_spread_of_future_rev_start =
           tokenptr->prev_space_spread;

        tokenptr = tmptr;
	if (tokenptr != NULL) {
            if (tokenptr->ending)
               no_of_words++;
            if (stretchable_fonts[tokenptr->font])
               at_least_one_char_in_stretchable_font = TRUE;
	}
      }

      if (original_prev_of_future_rev_end_is_a_begining) {
	 rev_start->begining = NOT_BEGIN;
	 rev_start->prev_space_width = 0;
	 rev_start->prev_space_spread = 0;
      }
      /* the following must be after the above to override the above */
      if (future_rev_end_is_now_a_begining) {
	 rev_start->begining = BEGINING;
	 rev_start->prev_space_width =
             original_prev_space_width_of_future_rev_end;
	 rev_start->prev_space_spread =
             original_prev_space_spread_of_future_rev_end;
      } else {
	 rev_start->begining = NOT_BEGIN;
	 rev_start->prev_space_width = 0;
	 rev_start->prev_space_spread = 0;
      }
      if (original_next_of_future_rev_start_is_now_an_ending) {
	 rev_end->ending = NOT_END;
	 rev_end->next_space_width = 0;
	 rev_end->next_space_spread = 0;
      }
      /* the following must be after the above to override the above */
      if (future_rev_start_is_now_an_ending) {
	 rev_end->ending = END;
	 rev_end->next_space_width =
           original_next_space_width_of_future_rev_start;
	 rev_end->next_space_spread =
           original_next_space_spread_of_future_rev_start;
      } else {
	 rev_end->ending = NOT_END;
	 rev_end->next_space_width = 0;
	 rev_end->next_space_spread = 0;
      }

      if (continue_from==NULL) {
         *start = rev_start;
      } else {
         continue_from->next = rev_start;
      }
      if (*end==rev_start) {
         *end = rev_end;
      }
      rev_end->next = tokenptr;

/*       printf("\nbefore recalculate_horizontal:\n");
         print_line(start,end);
         printf("sub_line:\n");
         print_line(&rev_start,&rev_end); */

      recalculate_horizontal(rev_start,rev_end);

/*       printf("\nafter recalculate_horizontal:\n");
         print_line(start,end);
         printf("sub_line:\n");
         print_line(&rev_start,&rev_end); */

    }
  }

  /* stretch Arabic words and reposition letters .. Johny */
  if ((stretch_mode != NONE || stretch_places != NONE) &&
      (no_stretchable_fonts > 0 &&
       at_least_one_char_in_stretchable_font) ){
      total_stretch = calculate_total_stretching(*start,*end);

/*    printf("\nfinal total_stretch: %d\n", total_stretch);
      printf("global no_of_words: %d\n", no_of_words);
      printf("global Number_of_Words procedure: %d\n",
           Number_of_Words(*start,*end));   
      printf("at_least_one_char_in_stretchable_font: %d\n",
           at_least_one_char_in_stretchable_font);    */

      if (total_stretch < 0)       /* the line was shrunk!! */
          total_stretch = 0;
      if (total_stretch > 0 ) {
          find_all_stretching_places(*start,*end,stretch_mode);


          if (no_stretching_places > 0) {
/*            printf("\nbefore stretch_letters\n");
              print_line(start,end); */
              stretch_letters(*start,*end,total_stretch,stretch_mode);
/*            printf("\nafter stretch_letters\n");
              print_line(start,end); */
          }
      }
  }

/*=================*/
  slide_connection_stretches_through_diacriticals_to_proper_token(*start,*end);
  if (any_diacriticals) {
      move_diacriticals_to_ends_of_words(start,end);
  }
/*=================*/
  tokenptr = *start;      /* print tokens of line */
  while (tokenptr != NULL)
  {
    put_token(tokenptr,start_of_line);
    start_of_line = FALSE;
    tokenptr = tokenptr->next;
  }

/* printf("\nat end:\n");
   print_line(start,end); */

  free_line(start,end);
  
  return;
}

/*************************************************************************
**************************************************************************
**                                                                      **
** this routine reverses the specified internal line                    **
** It also recognizes ZERO width characters and preserves their         **
** order with the next letter.                                          **
**                                                                      **
** Johny                                                                **
**************************************************************************
*************************************************************************/


reverse_line(start,end,paper_width)

TOKENPTR *start;
TOKENPTR *end;
int      paper_width;

{
  int       line_length = (*end)->horizontal_pos + (*end)->width;

/* PERHAPS THE -1 ON THE NEXT LINE SHOULD BE ELIMINATED TO GET THE LEFT
   MARGIN TO REMAIN AT 576 INSTEAD OF 575!!! */

  int       new_indent  = paper_width - line_length - 1;
  int       adjustment  = new_indent - (*start)->horizontal_pos;
  TOKENPTR  tmptr = (*start);
  TOKENPTR  save_start;
  TOKENPTR  tmp_start   = NULL;
  TOKENPTR  tmp_end     = NULL;
  bool      begining;
  bool      ending;
  int next_space_width;
  int prev_space_width;
  int next_space_spread;
  int prev_space_spread;
  int       total_stretch;
  TOKENPTR  zerowc[MAXZWC];
  int       i,j,hpos;

/* printf("reverse_line: %d %d %d\n", start, end, paper_width);
   printf("before reversing:\n");
   print_line(start,end); */

  while ((tmptr = (*start)) != NULL)
  {
    *start = (*start)->next;

    begining = tmptr->begining;
    ending   = tmptr->ending;
    next_space_width = tmptr->next_space_width;
    prev_space_width = tmptr->prev_space_width;
    next_space_spread = tmptr->next_space_spread;
    prev_space_spread = tmptr->prev_space_spread;
    tmptr->begining = ending;
    tmptr->ending   = begining;
    tmptr->next_space_width = prev_space_width;
    tmptr->prev_space_width = next_space_width;
    tmptr->next_space_spread = prev_space_spread;
    tmptr->prev_space_spread = next_space_spread;
    hpos = tmptr->horizontal_pos;
    i = 0;
        /* enter the equaly hpos tokens into an array */
    while ((tmptr->next != NULL) && (tmptr->next->horizontal_pos == hpos))
    {
        zerowc[i++] = tmptr;
        tmptr = tmptr->next;
    }
    if (i>0) {
        int new_first_zw_begining;
        int new_first_zw_prev_space_width;
        int new_first_zw_prev_space_spread;

	any_diacriticals = TRUE;
        *start = tmptr->next;
        zerowc[i] = tmptr;
        /* update the horizontal pos for each one of them */
        for (j=i; j>=0; j--) {
            zerowc[j]->horizontal_pos += adjustment;
            push_token(zerowc[j],&tmp_start,&tmp_end);
        }
        new_first_zw_begining = zerowc[i]->ending;
        new_first_zw_prev_space_width = zerowc[i]->next_space_width;
        new_first_zw_prev_space_spread = zerowc[i]->next_space_spread;

        zerowc[i]->ending = zerowc[0]->ending;
        zerowc[i]->next_space_width = zerowc[0]->next_space_width;
        zerowc[i]->next_space_spread = zerowc[0]->next_space_spread;

        zerowc[0]->ending = NOT_END;
        zerowc[0]->next_space_width = 0;
        zerowc[0]->next_space_spread = 0;

        zerowc[0]->begining = new_first_zw_begining;
        zerowc[0]->prev_space_width = new_first_zw_prev_space_width;
        zerowc[0]->prev_space_spread = new_first_zw_prev_space_spread;
    }
    else if (tmptr != NULL) {
        tmptr->horizontal_pos += adjustment;
        push_token(tmptr,&tmp_start,&tmp_end);
    }
  }

  *start = tmp_start;
  *end   = tmp_end;

  recalculate_horizontal(*start,*end);
  
  return;
}

/************************************************************************* 
************************************************************************** 
**                                                                      **
** this routine recalculates the horizontal motion in the specified     **
** internal line                                                        **
**                                                                      **
************************************************************************** 
*************************************************************************/ 

recalculate_horizontal(start,end)

  TOKENPTR  start;                  /* ptr to start of line */
  TOKENPTR  end  ;                  /* ptr to end of line   */

{

  TOKENPTR curr_token;
  TOKENPTR next_token;
  int      next_horizontal;
  int      curr_horizontal;

/* printf("recalculate_horizontal: %d %d\n", start, end);
   printf("before recalculation:\n");
   print_line(&start,&end); */

  curr_token = start; 
  curr_horizontal = end->horizontal_pos;
  curr_token->horizontal_pos = curr_horizontal;

  while ( curr_token != end ) {
      next_token = curr_token->next;
      next_horizontal = curr_horizontal + curr_token->width +
          curr_token->next_space_width + curr_token->next_space_spread;
      curr_horizontal = next_horizontal;
      curr_token = next_token;
      curr_token->horizontal_pos = curr_horizontal;
  }
  
/* printf("after recalculation:\n");
   print_line(&start,&end); */

  return;
}

/*************************************************************************
**************************************************************************
**                                                                      **
** This function should return the total stretching amount.             **
** It does that, by calculating the total amount of extra interword     **
** spaces.
**                                                                      **
** Johny                                                                **
**************************************************************************
*************************************************************************/

int
calculate_total_stretching (start,end)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */

{
  TOKENPTR tokenptr  = start;
  int      total_stretch = 0;

/* fprintf(stderr, "calculate_total_stretching: %d %d\n", start, end); */
/* printf("total_stretch: %d\n",total_stretch);   */

  while (tokenptr != NULL) {
     if (tokenptr->ending) {
        total_stretch += tokenptr->next_space_spread;

/*      printf("total_stretch: %d\n",total_stretch); */

     }
     tokenptr = tokenptr->next;
  }
  return(total_stretch);
}

/*************************************************************************
**************************************************************************
**                                                                      **
** find all possible stretching places based on the stretching mode.    **
** stretch_mode can be one of the following:				**
**	     NONE : dont stretch at all.				**
**	   LETTER : stretch stretchable letters internally              **
**     CONNECTION : stretch connections to connect_before and connect-  **
**		    both letters                                        **
**	   EITHER : connect either a stretchable letter or a connection **
**		    whichever is later in the word                      **
**	     BOTH : stretch both the letter and the connection to the   **
**		    letter if it is the last connectable place within   **
**		    a word otherwise as EITHER                          **
**									**
** This routine finds at most one token for stretching per word,        **
** according to the mode.                                               **
**									**
** Note that the characters are already in printing order. Therefore,   **
** the last characters in Arabic or Persian words are the first in the  **
** words in the scanning order.                                         **
**                                                                      **
**************************************************************************
*************************************************************************/

find_all_stretching_places(start,end,mode)
        TOKENPTR start;                  /* ptr to start of line */
        TOKENPTR end;                    /* ptr to end of line   */
        char mode;			 /* stretching mode      */
{
TOKENPTR current_token = start;

no_stretching_places = 0;
no_connection_stretching_places = 0;
no_letter_stretching_places = 0;

/* fprintf(stderr,"find_all_stretching_places: %d %d %d\n", start,end,mode); */
while (current_token != NULL) {
        /* skip non arabic/persian characters */
        while ((current_token != NULL) &&
                        (!stretchable_fonts[current_token->font])){
                current_token = current_token->next;
        }
        if (current_token != NULL){
                if (no_add_to_manually_stretched_word){
                    TOKENPTR search_token=current_token;
                    bool found_a_manual_stretch=FALSE;
                    bool have_searched_entire_word=FALSE;
                    while (search_token != NULL &&
                          !have_searched_entire_word &&
                          !found_a_manual_stretch ){
                        if ( (search_token->already_stretched_amount > 0) ||
                             (search_token->
                                 already_stretched_A_connection_amount > 0) ||
                             (search_token->
                                 already_stretched_B_connection_amount > 0) ||
                             ((search_token->char1 == 'h') &&
                              (search_token->char2 == 'y')) ||
                             ((search_token->char1 == 'q') &&
                              (search_token->char2 == '8') &&
                              (search_token->char3 == '3')) ){
                            found_a_manual_stretch=TRUE;
                        }
                        if (search_token->ending){
                            have_searched_entire_word=TRUE;
                        } else {
                            search_token = search_token->next;
                        }
                    }
                    /* search_token==NULL || have_searched_entire_word ||
                       found_a_manual_stretch */
                    if (search_token==NULL) goto end_of_loop;
                    if (found_a_manual_stretch){
                        if (have_searched_entire_word){
                            current_token = search_token;
                        } else {
                            current_token = search_token;
                            while ((current_token != NULL) &&
                                   !current_token->ending){
                                current_token = current_token->next;
                            }
                        }
                    }
                }
                switch(mode){
                case NONE:{
                        break;
                }
                case LETTER:{
                        if(
                            (current_token->internal_stretchability == 'S') &&
                            ((get_stretching_mode(current_token->font)==
                                LETTER) ||
                             (get_stretching_mode(current_token->font)==BOTH))){
                            if(
                               (!no_add_to_manually_stretched_letter
                              ||
                               (no_add_to_manually_stretched_letter &&
                                current_token->already_stretched_amount
                                 == 0))
                            ){
                                stretching_places[no_stretching_places]=
                                        current_token;
                                letter_stretching_place_index[
                                        no_letter_stretching_places]=
                                        no_stretching_places;
                                no_letter_stretching_places++;
                                stretching_style[no_stretching_places]=LETTER;
                                no_stretching_places++;
                                while ((current_token != NULL) &&
                                                !(current_token->ending)){
                                        current_token = current_token->next;
                                }
                            }
                        }
                        break;
                }
                case CONNECTION:{
                        if (
                            (current_token->connectivity == 'B' ||
                             current_token->connectivity == 'P') &&
                            ((get_stretching_mode(current_token->font)==
                                  CONNECTION) ||
                             (get_stretching_mode(current_token->font)==BOTH))){

                            if(
                                (!no_add_to_manually_stretched_connection
                                ||
                                (no_add_to_manually_stretched_connection
                                &&
                                (!(current_token->next->char1 == 'h' &&
                                   current_token->next->char2 == 'y' &&
                                   current_token->next->char3 =='\0')&&
                                 !(current_token->next->char1 == '1' &&
                                   current_token->next->char2 == '8' &&
                                   current_token->next->char3 == '3') &&
                                 (current_token->
                                     already_stretched_B_connection_amount==0))
                           ))){
                                stretching_places[no_stretching_places]=
                                        current_token;
                                connection_stretching_place_index[
                                        no_connection_stretching_places]=
                                        no_stretching_places;
                                no_connection_stretching_places++;
                                stretching_style[no_stretching_places]=
                                        CONNECTION;
                                no_stretching_places++;
                                while ((current_token != NULL) &&
                                                !(current_token->ending)){
                                        current_token = current_token->next;
                                }
                            }
                        }
                        break;
                }
                case EITHER:{
                        if(
                           ((current_token->internal_stretchability == 'S') &&
                             ((get_stretching_mode(current_token->font)==
                                   LETTER) ||
                              (get_stretching_mode(current_token->font)==BOTH))
                           ) &&
                            (!no_add_to_manually_stretched_letter
                           ||
                            (no_add_to_manually_stretched_letter &&
                             current_token->already_stretched_amount == 0))){
                                stretching_places[no_stretching_places]=
                                        current_token;
                                letter_stretching_place_index[
                                        no_letter_stretching_places]=
                                        no_stretching_places;
                                no_letter_stretching_places++;
                                stretching_style[no_stretching_places]=LETTER;
                                no_stretching_places++;
                                while ((current_token != NULL) &&
                                                !(current_token->ending)){
                                        current_token = current_token->next;
                                }
                        } else if (((current_token->connectivity == 'B' ||
                                     current_token->connectivity == 'P') &&
                                    ((get_stretching_mode(current_token->font)==
                                          CONNECTION) ||
                                     (get_stretching_mode(current_token->font)==
                                          BOTH))) &&
                                   (!no_add_to_manually_stretched_connection
                                   ||
                                   (no_add_to_manually_stretched_connection
                                   &&
                                   (!(current_token->next->char1 == 'h' &&
                                      current_token->next->char2 == 'y' &&
                                      current_token->next->char3 =='\0')&&
                                    !(current_token->next->char1 == '1' &&
                                      current_token->next->char2 == '8' &&
                                      current_token->next->char3 == '3') &&
                                    (current_token->
                                        already_stretched_B_connection_amount==
                                            0))))){
                                stretching_places[no_stretching_places]=
                                        current_token;
                                connection_stretching_place_index[
                                        no_connection_stretching_places]=
                                        no_stretching_places;
                                no_connection_stretching_places++;
                                stretching_style[no_stretching_places]=
                                        CONNECTION;
                                no_stretching_places++;
                                while ((current_token != NULL) &&
                                                !(current_token->ending)){
                                        current_token = current_token->next;
                                }
                        }
                        break;
                }
                case BOTH:{
                        if ((current_token->internal_stretchability == 'S') &&
                            ((get_stretching_mode(current_token->font)==
                                  LETTER) ||
                             (get_stretching_mode(current_token->font)==BOTH))){
                                if ((current_token->connectivity == 'B' ||
                                     current_token->connectivity == 'P') &&
                                    ((get_stretching_mode(current_token->font)==
                                          CONNECTION) ||
                                     (get_stretching_mode(current_token->font)==
                                          BOTH))){
                                        if (
                                         (!no_add_to_manually_stretched_letter
                                         ||
                                         (no_add_to_manually_stretched_letter &&
                                         current_token->already_stretched_amount
                                         == 0))
                                        &&
                                        (!no_add_to_manually_stretched_connection
                                        ||
                                        (no_add_to_manually_stretched_connection
                                         &&
                                         (!(current_token->next->char1 == 'h' &&
                                            current_token->next->char2 == 'y' &&
                                            current_token->next->char3 =='\0')&&
                                          !(current_token->next->char1 == '1' &&
                                            current_token->next->char2 == '8' &&
                                            current_token->next->char3 == '3')&&
                                          (current_token->
                                           already_stretched_B_connection_amount
                                           ==0)
                                         )))){
                                                stretching_places[
                                                    no_stretching_places]=
                                                    current_token;
                                                letter_stretching_place_index[
                                                    no_letter_stretching_places]=
                                                    no_stretching_places;
                                                no_letter_stretching_places++;
                                                stretching_style[
                                                    no_stretching_places]=
                                                    LETTER;
                                                no_stretching_places++;

                                                /* putting this token in a 2nd
                                                time so that it can be
                                                stretched both in the
                                                connection and in the letter */
        
                                                stretching_places[
                                                    no_stretching_places]=
                                                    current_token;
                                                connection_stretching_place_index[
                                                    no_connection_stretching_places]=
                                                    no_stretching_places;
                                                no_connection_stretching_places++;
                                                stretching_style[
                                                    no_stretching_places]=
                                                    CONNECTION;
                                                no_stretching_places++;
                                                while((current_token != NULL) &&
                                                    !(current_token->ending)){
                                                        current_token =
                                                           current_token->next;
                                                }
                                        }
                                } else {
                                        if (
                                         !no_add_to_manually_stretched_letter ||
                                         (no_add_to_manually_stretched_letter &&
                                         current_token->already_stretched_amount
                                         == 0)) {
                                                stretching_places[
                                                    no_stretching_places]=
                                                    current_token;
                                                letter_stretching_place_index[
                                                    no_letter_stretching_places]=
                                                    no_stretching_places;
                                                no_letter_stretching_places++;
                                                stretching_style[
                                                    no_stretching_places]=
                                                    LETTER;
                                                no_stretching_places++;
                                                while((current_token != NULL) &&
                                                    !(current_token->ending)){
                                                        current_token =
                                                           current_token->next;
                                                }
                                        }
                                }
                        } else if ((current_token->connectivity == 'B' ||
                                    current_token->connectivity == 'P') &&
                                   ((get_stretching_mode(current_token->font)==
                                         CONNECTION) ||
                                    (get_stretching_mode(current_token->font)==
                                         BOTH))){
                                if ( !no_add_to_manually_stretched_connection ||
                                     (no_add_to_manually_stretched_connection &&
                                         (!(current_token->next->char1 == 'h' &&
                                            current_token->next->char2 == 'y' &&
                                            current_token->next->char3 =='\0')&&
                                          !(current_token->next->char1 == '1' &&
                                            current_token->next->char2 == '8' &&
                                            current_token->next->char3 == '3')&&
                                          (current_token->
                                           already_stretched_B_connection_amount
                                           ==0)))){
                                        stretching_places[no_stretching_places]=
                                                current_token;
                                        connection_stretching_place_index[
                                                no_connection_stretching_places]=
                                                no_stretching_places;
                                        no_connection_stretching_places++;
                                        stretching_style[no_stretching_places]=
                                                CONNECTION;
                                        no_stretching_places++;
                                        while((current_token != NULL) &&
                                            !(current_token->ending)){
                                                current_token =
                                                   current_token->next;
                                        }
                                }
                        }
                        break;
                }
                }
        }
        if (current_token != NULL) {
                current_token = current_token->next;
        }
}
end_of_loop:; /* is this the right place */
}

/*************************************************************************
**************************************************************************
**                                                                      **
** stretch the appropriate letters, based on the stretching places.     **
** stretch_places can be one of the following:                          **
**       NONE : dont stretch at all.                                    **
**      FINAL : stretch the final word only.                            **
**     FINAL2 : stretch the final word only when mode is BOTH.          **
**     TO_MAX : stretch the last word(s). the difference from the       **
**              FINAL place, is when the stretching amount is too       **
**              big for one word ... in that case, the remainder        **
**              will be given to the previous one.                      **
**        ALL : stretch all of the Arabic, Hebrew, and Persian          **
**                  words in the line, equally.                         **
**                                                                      **
** Note that the characters are already in printing order. Therefore,   **
** the last characters in Arabic, Hebrew or Persian words are the       **
** first in the words in the scanning order.                            **
**                                                                      **
**************************************************************************
*************************************************************************/

int stretch_letters (start,end,total_stretch,mode)
        TOKENPTR start;                  /* ptr to start of line */
        TOKENPTR end;                    /* ptr to end of line   */
        int total_stretch;
        char mode;                   /* stretching mode      */
{
int leftover = 0;

/* fprintf(stderr, "stretch_letters: %d %d %d %d\n", start, end,
       total_stretch, mode); */

switch (stretch_places) {
        case NONE: {
                leftover=total_stretch;
                break;
        }

        case FINAL: {
                int stretch_amount;
                int filler_width;
                int no_fillers;
                if (stretching_style[0]==CONNECTION) {
                        if (get_connection_stretching_mode(
                                        stretching_places[0]->font)==
                                        FIXED_FILLER) {
                                filler_width =
                                        width2("hy",
                                        stretching_places[0]->point_size,
                                        stretching_places[0]->font);
                                no_fillers = total_stretch/filler_width;
                                stretch_amount = no_fillers*filler_width;
                                leftover = total_stretch%filler_width;
                                stretching_places[0]->fillers_num=no_fillers;
                                stretching_places[0]->filler_width=filler_width;
                        } else if (get_connection_stretching_mode(
                                        stretching_places[0]->font)==
                                        STRETCHABLE_FILLER) {
                                stretching_places[0]->filler_stretch =
                                        total_stretch;
                                leftover=0;
                        } else {
                                stretching_places[0]->connection_stretch =
                                        total_stretch;
                                stretching_places[0]->width +=
                                        total_stretch;
                                leftover=0;
                        }
                } else {
                        stretching_places[0]->letter_stretch= total_stretch;
                        stretching_places[0]->width+= total_stretch;
                        leftover = 0;
                }
                break;
        }
        case FINAL2: {
                int stretch_amount;
                int filler_width;
                int no_fillers;
                int connection_stretch;
                int letter_stretch;

                if (mode != BOTH) {
                        error(FATAL,
                            "FINAL2 places available only with BOTH mode\n");
                } else if ((no_stretching_places >= 2) && 
                           /* and they are in the same token */
                           (stretching_places[0] == stretching_places[1]) &&
                           (stretching_style[0] == LETTER) &&
                           (stretching_style[1] == CONNECTION)) {
                        stretch_amount = (int)(((double)total_stretch /2.0)+.5);
                        if (get_connection_stretching_mode(
                                        stretching_places[1]->font)==
                                        FIXED_FILLER) {
                                filler_width =
                                        width2("hy",
                                        stretching_places[0]->point_size,
                                        stretching_places[0]->font);
                                no_fillers = stretch_amount/filler_width;
                                connection_stretch = no_fillers*filler_width;
                                letter_stretch =
                                        total_stretch-connection_stretch;

                                stretching_places[0]->letter_stretch=
                                                letter_stretch;
                                stretching_places[0]->width+=
                                                letter_stretch;

                                stretching_places[1]->fillers_num=no_fillers;
                                stretching_places[1]->filler_width=filler_width;

                                leftover=0;
                        } else if (get_connection_stretching_mode(
                                        stretching_places[1]->font)==
                                        STRETCHABLE_FILLER) {
                                stretching_places[0]->letter_stretch =
                                        stretch_amount;
                                stretching_places[0]->width +=
                                        stretch_amount;
                                stretching_places[1]->filler_stretch =
                                        total_stretch - stretch_amount;
                                leftover=0;
                        } else {
                                stretching_places[0]->letter_stretch =
                                        stretch_amount;
                                stretching_places[0]->width +=
                                        stretch_amount;
                                stretching_places[1]->connection_stretch =
                                        total_stretch - stretch_amount;
                                stretching_places[1]->width +=
                                        total_stretch - stretch_amount;
                                leftover=0;
                        }
                } else /* know that (no_stretching_places == 1) */ {
                        if (stretching_style[0]==CONNECTION) {
                                if (get_connection_stretching_mode(
                                                stretching_places[0]->font)==
                                                FIXED_FILLER) {
                                        filler_width =
                                            width2("hy",
                                            stretching_places[0]->point_size,
                                            stretching_places[0]->font);
                                        no_fillers = total_stretch/filler_width;
                                        stretch_amount =
                                                no_fillers*filler_width;
                                        leftover = total_stretch%filler_width;
                                        stretching_places[0]->fillers_num=
                                                no_fillers;
                                        stretching_places[0]->filler_width=
                                                filler_width;
                                } else if (get_connection_stretching_mode(
                                                stretching_places[0]->font)==
                                                STRETCHABLE_FILLER) {
                                        stretching_places[0]->
                                                filler_stretch =
                                                total_stretch;
                                } else {
                                        stretching_places[0]->
                                                connection_stretch =
                                                total_stretch;
                                        stretching_places[0]->
                                                width += total_stretch;
                                        leftover=0;
                                }
                        } else {
                                stretching_places[0]->letter_stretch=
                                        total_stretch;
                                stretching_places[0]->width+=
                                        total_stretch;
                                leftover = 0;
                        }
                }
                break;
        }
        case TO_MAX: {
                int current_point_size;
                int limit_units;
                int current_place;
                int stretch_left;
                int filler_width;
                int no_fillers;
                int stretch_amount;

                current_point_size = stretching_places[0]->point_size;
                /* recall that \| is 1/6 em */
                limit_units =
                        stretch_limit*
                        6*width2("\\|",stretching_places[0]->point_size,
                                stretching_places[0]->font);
                current_place = 1;
                stretch_left = total_stretch;

                while ((no_stretching_places >= current_place) &&
                        stretch_left > 0) {
                        if (stretch_left > limit_units) {
                                if (stretching_style[current_place-1]==
                                                CONNECTION) {
                                        if (get_connection_stretching_mode(
                                                        stretching_places[
                                                        current_place-1]->font)
                                                        ==FIXED_FILLER) {
                                                filler_width=
                                                        width2("hy",
                                                        stretching_places[
                                                            current_place-1]->
                                                            point_size,
                                                        stretching_places[
                                                            current_place-1]->
                                                            font);
                                                no_fillers=limit_units/
                                                        filler_width;
                                                stretch_amount=no_fillers*
                                                        filler_width;
                                                stretch_left+=limit_units%
                                                        filler_width;
                                                stretching_places[
                                                        current_place-1]->
                                                        fillers_num=no_fillers;
                                                stretching_places[
                                                        current_place-1]->
                                                        filler_width=
                                                        filler_width;
                                        }else if(get_connection_stretching_mode(
                                                        stretching_places[
                                                        current_place-1]->font)
                                                        ==STRETCHABLE_FILLER) {
                                                stretching_places[
                                                        current_place-1]->
                                                        filler_stretch=
                                                        limit_units;
                                        } else {
                                                stretching_places[
                                                        current_place-1]->
                                                        connection_stretch=
                                                        limit_units;
                                                stretching_places[
                                                        current_place-1]->
                                                        width += limit_units;
                                        }
                                } else {
                                        stretching_places[current_place-1]->
                                                letter_stretch=limit_units;
                                        stretching_places[current_place-1]->
                                                width+= limit_units;
                                }
                                stretch_left -= limit_units;
                        } else {
                                if (stretching_style[current_place-1]==
                                                CONNECTION) {
                                        if (get_connection_stretching_mode(
                                                        stretching_places[
                                                        current_place-1]->font)
                                                        ==FIXED_FILLER) {
                                                filler_width=
                                                        width2("hy",
                                                        stretching_places[
                                                            current_place-1]->
                                                            point_size,
                                                        stretching_places[
                                                            current_place-1]->
                                                            font);
                                                no_fillers=stretch_left/
                                                        filler_width;
                                                stretch_amount=no_fillers*
                                                        filler_width;
                                                leftover+=stretch_left%
                                                        filler_width;
                                                stretching_places[
                                                        current_place-1]->
                                                        fillers_num=no_fillers;
                                                stretching_places[
                                                        current_place-1]->
                                                        filler_width=
                                                        filler_width;
                                        }else if(get_connection_stretching_mode(
                                                        stretching_places[
                                                        current_place-1]->font)
                                                        ==STRETCHABLE_FILLER) {
                                                stretching_places[
                                                        current_place-1]->
                                                        filler_stretch=
                                                        stretch_left;
                                        } else {
                                                stretching_places[
                                                        current_place-1]->
                                                        connection_stretch=
                                                        stretch_left;
                                                stretching_places[
                                                        current_place-1]->
                                                        width += stretch_left;
                                        }
                                } else {
                                        stretching_places[current_place-1]->
                                                letter_stretch=stretch_left;
                                        stretching_places[current_place-1]->
                                                width += stretch_left;
                                }
                                stretch_left = 0;
                        }
                        current_place++;
                }
                /* we know that no_stretching_places < current_place OR
                stretch_left == 0 */
                leftover += stretch_left;
                break;
        }
        case ALL: {
                int total_connections_stretch;
                int total_letters_stretch;
                int uniform_stretch_per_connection;
                int filler_width[MAXSTRETCHINGPLACES];
                int fillers_per_connection[MAXSTRETCHINGPLACES];
                int stretch_per_connection[MAXSTRETCHINGPLACES];
                int stretch_per_letter;
                int i;
                int index;
                int no_new_fillers;
                TOKENPTR tokenptr;

                total_connections_stretch=(int)((
                        (double)no_connection_stretching_places/
                        (double)no_stretching_places)*total_stretch);
                total_letters_stretch=(int)((
                        (double)no_letter_stretching_places/
                        (double)no_stretching_places)*total_stretch);
                leftover=total_stretch - total_connections_stretch - 
                        total_letters_stretch;

                if (no_connection_stretching_places >= 1){

                        uniform_stretch_per_connection=
                                total_connections_stretch/
                                no_connection_stretching_places;
                        leftover+=(total_connections_stretch%
                                no_connection_stretching_places);
                        for (i=0; i<=no_connection_stretching_places-1; i++){
                                index=
                                    connection_stretching_place_index[
                                        i];
                                tokenptr=stretching_places[index];
                                if (get_connection_stretching_mode(
                                                tokenptr->font)==FIXED_FILLER){
                                        filler_width[i]=
                                                width2("hy",
                                                tokenptr->point_size,
                                                tokenptr->font);
                                        fillers_per_connection[i]=
                                                uniform_stretch_per_connection/
                                                filler_width[i];
                                        stretch_per_connection[i]=
                                                fillers_per_connection[i]*
                                                filler_width[i];
                                        leftover+=
                                                (uniform_stretch_per_connection%
                                                filler_width[i]);
                                        stretching_places[index]->fillers_num=
                                                fillers_per_connection[i];
                                        stretching_places[index]->filler_width=
                                                filler_width[i];
                                } else if (get_connection_stretching_mode(
                                                tokenptr->font)==
                                                STRETCHABLE_FILLER){
                                        tokenptr->filler_stretch=
                                                uniform_stretch_per_connection;
                                        /*leave leftover alone*/
                                } else {
                                        tokenptr->connection_stretch=
                                                uniform_stretch_per_connection;
                                        tokenptr->width +=
                                                uniform_stretch_per_connection;
                                        /*leave leftover alone*/
                                }
                        }

                } else {
                        uniform_stretch_per_connection=0;
                }
                if (no_letter_stretching_places >= 1){
                        total_letters_stretch+=leftover;
                        stretch_per_letter=total_letters_stretch/
                                no_letter_stretching_places;
                        /* all previous leftovers moved to letters
                        stretch, so now leftovers should be 0,
                        so no +=.. just = */
                        leftover=(total_letters_stretch%
                                no_letter_stretching_places);
                } else {
                        stretch_per_letter=0;
                }
                if (favor_stretch) {
                        if (mode != CONNECTION){
                                if (no_letter_stretching_places >= 1){
                                        stretch_per_letter+=leftover/
                                                no_letter_stretching_places;
                                        leftover=leftover%
                                                no_letter_stretching_places;
                                }
                        }
                } /* else favor distribution 
                        and do nothing */
                for (i=0; i <= no_stretching_places-1; i++){
                        if (stretching_style[i]==LETTER){
                                stretching_places[i]->letter_stretch +=
                                        stretch_per_letter;
                                stretching_places[i]->width +=
                                        stretch_per_letter;
                        } /* else stretching_style[i]==CONNECTION
                                and stretching_place[i] was already
                                stretched above */
                }
                break;
        }
}
recalculate_spaces(start,end,leftover);
}

/*************************************************************************
**************************************************************************
**                                                                      **
** recalculates the inter words spaces, after stretching a specified    **
** letter. It simply removes the extra spaces added by troff            **
** and spreads the amount of leftover, which is basically the           **
** remainder after swallowing up the extra spaces by an integral        **
** number of fillers                                                    **
**                                                                      **
** Johny.                                                               **
**                                                                      **
**************************************************************************
*************************************************************************/

recalculate_spaces (start,end,leftover)
        TOKENPTR start;                  /* ptr to start of line */
        TOKENPTR end  ;                  /* ptr to end of line   */
        int leftover;
{
TOKENPTR tokenptr;
TOKENPTR nextptr;

int horizontal_pos;
int extra_per_gap;
int remainder;
int no_of_gaps;
int i;

/* fprintf(stderr, "recalculate_spaces: %d %d %d\n", start, end, leftover); */

if (no_of_words == 1) return; 
no_of_gaps = no_of_words - 1;
extra_per_gap = leftover / no_of_gaps;
remainder  = (leftover % no_of_gaps);

left_to_right_sweep = !left_to_right_sweep;
tokenptr = start;
horizontal_pos = tokenptr->horizontal_pos;
i = 1;

while (tokenptr != NULL) {
        if ((tokenptr->ending) && ((nextptr=tokenptr->next) != NULL)) {
                if ((left_to_right_sweep && (i>(no_of_gaps-remainder)))
                        ||
                    (!left_to_right_sweep && i <= remainder)) {
                        tokenptr->next_space_spread = extra_per_gap + 1;
                        nextptr->prev_space_spread = extra_per_gap + 1;
                } else {
                        tokenptr->next_space_spread = extra_per_gap;
                        nextptr->prev_space_spread = extra_per_gap;
                }
        i++;
        }

        tokenptr->horizontal_pos = horizontal_pos;
        horizontal_pos +=
                tokenptr->width +
                tokenptr->next_space_width +
                tokenptr->next_space_spread +
                (tokenptr->fillers_num * tokenptr->filler_width) +
                tokenptr->filler_stretch;

        tokenptr = tokenptr->next;
}

}

/*************************************************************************
**************************************************************************
**                                                                      **
** return the number of words in the current line.                      **
** Johny.                                                               **
**                                                                      **
**************************************************************************
*************************************************************************/

int
Number_of_Words(start,end)

TOKENPTR  start;                  /* ptr to start of line */
TOKENPTR  end  ;                  /* ptr to end of line   */
{
  TOKENPTR  tokenptr = start;
  int       nw = 0;               /* the number of Arabic words in line */
/* fprintf(stderr, "Number_of_words: %d %d\n", start, end); */

  while (tokenptr != NULL) {
    if (tokenptr->ending)
        nw++;
    tokenptr = tokenptr->next;
  }
  return(nw);
}

/*************************************************************************
**************************************************************************
**                                                                      **
** reverse nested numbers                                               **
** Dan                                                                  **
**                                                                      **
**************************************************************************
*************************************************************************/

reverse_number(start,end,continue_from)
    TOKENPTR *start, *end, continue_from;
{
    TOKENPTR tmptr;
    TOKENPTR tokenptr;
    TOKENPTR rev_start = NULL;     /* ptr to start of reversed token line */
    TOKENPTR rev_end = NULL;       /* ptr to end of reversed token line   */
    bool original_prev_of_future_rev_end_is_a_begining;
    bool future_rev_end_is_now_a_begining;
    bool original_next_of_future_rev_start_is_now_an_ending;
    bool future_rev_start_is_now_an_ending;
    int original_prev_space_width_of_future_rev_end;
    int original_prev_space_spread_of_future_rev_end;
    int original_next_space_width_of_future_rev_start;
    int original_next_space_spread_of_future_rev_start;
    bool      begining;
    bool      ending;
    int next_space_width;
    int prev_space_width;
    int next_space_spread;
    int prev_space_spread;
    TOKENPTR  zerowc[MAXZWC];
    int       i,j,hpos;

/*     printf("\nbefore reverse_number:\n");
       printf("sub_line:\n");
       print_line(start,end); */

tokenptr = *start;

if (continue_from == NULL){
   original_prev_of_future_rev_end_is_a_begining = FALSE;
} else {
   original_prev_of_future_rev_end_is_a_begining =
      continue_from->begining;
}
future_rev_end_is_now_a_begining = tokenptr->begining;
original_prev_space_width_of_future_rev_end =
    tokenptr->prev_space_width;
original_prev_space_spread_of_future_rev_end =
    tokenptr->prev_space_spread;

while (tokenptr != NULL)
/* tokenptr is set to NULL when the end of the number has been processed 
whose direction = the document direction */
{
  /* invert the direction of the token so that it is considered to be in
  time order AFTER the reversal */
  tokenptr->lr = !tokenptr->lr;
  /* swap word endings and beginnings */ 
  begining = tokenptr->begining;
  ending   = tokenptr->ending;
  next_space_width = tokenptr->next_space_width;
  prev_space_width = tokenptr->prev_space_width;
  next_space_spread = tokenptr->next_space_spread;
  prev_space_spread = tokenptr->prev_space_spread;
  tokenptr->begining = ending;
  tokenptr->ending   = begining;
  tokenptr->next_space_width = prev_space_width;
  tokenptr->prev_space_width = next_space_width;
  tokenptr->next_space_spread = prev_space_spread;
  tokenptr->prev_space_spread = next_space_spread;
  hpos = tokenptr->horizontal_pos;
  i = 0;
  while ((tokenptr->next != NULL) &&
         (tokenptr->next->horizontal_pos == hpos))
  {
      zerowc[i++] = tokenptr;
      tokenptr = tokenptr->next;
  }
  if (tokenptr == *end) {
      tmptr = NULL;
  } else {
      tmptr = tokenptr->next;
  }
  if (i>0)        /* a ZERO-width character was detected */
  {
    int new_first_zw_begining;
    int new_first_zw_prev_space_width;
    int new_first_zw_prev_space_spread;
  
    zerowc[i] = tokenptr;
    for (j=i; j>=0; j--) {
      push_token(zerowc[j],&rev_start,&rev_end);
    }
    new_first_zw_begining = zerowc[i]->ending;
    new_first_zw_prev_space_width = zerowc[i]->next_space_width;
    new_first_zw_prev_space_spread = zerowc[i]->next_space_spread;
  
    zerowc[i]->ending = zerowc[0]->ending;
    zerowc[i]->next_space_width = zerowc[0]->next_space_width;
    zerowc[i]->next_space_spread = zerowc[0]->next_space_spread;
  
    zerowc[0]->ending = NOT_END;
    zerowc[0]->next_space_width = 0;
    zerowc[0]->next_space_spread = 0;
  
    zerowc[0]->begining = new_first_zw_begining;
    zerowc[0]->prev_space_width = new_first_zw_prev_space_width;
    zerowc[0]->prev_space_spread = new_first_zw_prev_space_spread;
  }
  else push_token(tokenptr,&rev_start,&rev_end);
  
  /* doing now each time through the loop in case we exit at the next
  time at the top */
  if (tmptr == NULL){
     original_next_of_future_rev_start_is_now_an_ending = FALSE;
  } else {
     original_next_of_future_rev_start_is_now_an_ending = tmptr->ending;
  }
  /* on the assumption that x->begining and x->ending AND
  x->next_space_spread and x->prev_space_spread,
  x->next_space_width and x->prev_space_width, resp., for x in rev_end
  .. rev_start, have already been swapped */
  future_rev_start_is_now_an_ending = tokenptr->begining;
  original_next_space_width_of_future_rev_start =
     tokenptr->prev_space_width;
  original_next_space_spread_of_future_rev_start =
     tokenptr->prev_space_spread;
  
  tokenptr = tmptr;
}
  
if (original_prev_of_future_rev_end_is_a_begining) {
   rev_start->begining = NOT_BEGIN;
   rev_start->prev_space_width = 0;
   rev_start->prev_space_spread = 0;
}
/* the following must be after the above to override the above */
if (future_rev_end_is_now_a_begining) {
   rev_start->begining = BEGINING;
   rev_start->prev_space_width =
       original_prev_space_width_of_future_rev_end;
   rev_start->prev_space_spread =
       original_prev_space_spread_of_future_rev_end;
} else {
   rev_start->begining = NOT_BEGIN;
   rev_start->prev_space_width = 0;
   rev_start->prev_space_spread = 0;
}
if (original_next_of_future_rev_start_is_now_an_ending) {
   rev_end->ending = NOT_END;
   rev_end->next_space_width = 0;
   rev_end->next_space_spread = 0;
}
/* the following must be after the above to override the above */
if (future_rev_start_is_now_an_ending) {
   rev_end->ending = END;
   rev_end->next_space_width =
     original_next_space_width_of_future_rev_start;
   rev_end->next_space_spread =
     original_next_space_spread_of_future_rev_start;
} else {
   rev_end->ending = NOT_END;
   rev_end->next_space_width = 0;
   rev_end->next_space_spread = 0;
}
  
if (continue_from==NULL) {
   *start = rev_start;
} else {
   continue_from->next = rev_start;
}
if (*end == rev_start) {
   *end = rev_end;
}
rev_end->next = tokenptr;
  
/*     printf("\nbefore recalculate_horizontal:\n");
       printf("sub_line:\n");
       print_line(&rev_start,&rev_end); */
  
    recalculate_horizontal(rev_start,rev_end);
  
/*     printf("\nafter recalculate_horizontal:\n");
       printf("sub_line:\n");
       print_line(&rev_start,&rev_end);
       printf("\nand now again from the external pointers:\n");
       print_line(start,end); */
}

/*************************************************************************
**************************************************************************
**                                                                      **
** This routine calculates the length of the word beginning at the      **
** token pointed to by start.. skipping through tokens until one        **
** answering yes to ->ending is found. Note that if any zero-widths     **
** characters are found, adding their width of 0 to the accumulated     **
** length works just fine.                                              **
**                                                                      **
** Dan                                                                  **
**************************************************************************
*************************************************************************/

int Word_length(start)
TOKENPTR start;
{
	TOKENPTR tokenptr = start;
	int length = tokenptr->width;

	while ( (!(tokenptr->ending)) && (tokenptr != NULL)) {
		tokenptr = tokenptr->next;
		length += tokenptr->width;
	}
	return(length);
}

/*************************************************************************
**************************************************************************
**                                                                      **
** This routine slides the ACstretch requests through zero width        **
** characters (diacriticals) to first subsequent non-zero width char    **
**                                                                      **
** Dan                                                                  **
**************************************************************************
*************************************************************************/

slide_connection_stretches_through_diacriticals_to_proper_token(start,end)
	TOKENPTR start, end;
{
	TOKENPTR tokenptr, next_zero, next_non_zero;
	int later, at_this_time;
	tokenptr = start;
	while (tokenptr != NULL) {
		if (tokenptr->connection_stretch > 0) {
			next_non_zero = tokenptr->next;
			if (next_non_zero == NULL) {
				error(FATAL,
				"\nA connected-from letter is not preceeded by a letter\n");
			}
			while (next_non_zero->width == 0) {
				next_non_zero = next_non_zero->next;
				if (next_non_zero == NULL) {
					error(FATAL,
					"\nA connected-from letter is not preceeded by a letter\n");
				}
			}
			if (get_connection_stretching_mode(next_non_zero->font)
					== STRETCHABLE_CONNECTION) {
				at_this_time=(int)((double)
					(tokenptr->connection_stretch/2.0)+.5);
				later=tokenptr->connection_stretch-at_this_time;
				tokenptr->width -= later;
				tokenptr->already_stretched_B_connection_amount
					+=at_this_time;
				next_non_zero->
                                        already_stretched_A_connection_amount
					+=later;
/*				next_zero=tokenptr->next;
				while (next_zero != next_non_zero) {
					next_zero->horizontal_pos -= later;
					next_zero = next_zero->next;
				} */
				next_non_zero->horizontal_pos -= later;
			} else {
				tokenptr->already_stretched_B_connection_amount
						+=tokenptr->connection_stretch;
			}
		}
		tokenptr = tokenptr->next;
	}
	return;
}

/*************************************************************************
**************************************************************************
**                                                                      **
** This routine moves diacriticals to the end of the containing word    **
** so that they do not interrupt the contiguous flow of the non-        **
** diacriticals                                                         **
**                                                                      **
** Dan                                                                  **
**************************************************************************
*************************************************************************/

move_diacriticals_to_ends_of_words(start,end)
	TOKENPTR *start, *end;
{
	TOKENPTR tokenptr = *start;
	TOKENPTR prev = NULL;
	TOKENPTR start_of_diacriticals = NULL;
	TOKENPTR last_diacritical = NULL;

	while (tokenptr != NULL) {
		if (tokenptr->ending && (start_of_diacriticals != NULL)){
		/* if !tokenptr->begining then prev!=NULL */

			if (tokenptr->width != 0) {
			/* if tokenptr->begining then
			NULLity of prev is not relevant */

				if (tokenptr == *end) {
					*end = last_diacritical;
				}
				last_diacritical->next = tokenptr->next;
				tokenptr->next = start_of_diacriticals;
				tokenptr->ending = FALSE;
				last_diacritical->ending = TRUE;
				prev = last_diacritical;
				tokenptr = prev->next;

			} else {
			/* if tokenptr->begining then
			NULLity of prev is relevant */

				if (tokenptr->begining) {
					tokenptr->begining = FALSE;
					start_of_diacriticals->begining = TRUE;
				}
				last_diacritical->next = tokenptr;
				if (prev == NULL) {
					*start = start_of_diacriticals;
				} else {
					prev->next = start_of_diacriticals;
				}
				prev = tokenptr;
				tokenptr = tokenptr->next;
			}
			start_of_diacriticals = NULL;
			last_diacritical = NULL;

		} else if ( !tokenptr->ending && (tokenptr->width == 0)){

			if (prev == NULL) {
				*start = tokenptr->next;
			} else {
				prev->next=tokenptr->next;
			}

			if (prev == NULL && tokenptr->begining) {
				(*start)->begining = TRUE;
				tokenptr->begining = FALSE;
			}

			if (last_diacritical == NULL) {
				start_of_diacriticals = tokenptr;
				last_diacritical = tokenptr;
				tokenptr->next = NULL;
			} else {
				last_diacritical->next = tokenptr;
				last_diacritical = tokenptr;
				tokenptr->next = NULL;
			}

			if (prev == NULL) {
				tokenptr = *start;
			} else {
				tokenptr = prev->next;
			}
			/* leave prev where it was */

		} else {
			prev = tokenptr;
			tokenptr = tokenptr->next;
		}
	}
	return;
}

/* used for debugging ... Johny */

print_line(start,end)
TOKENPTR  *start;                 /* ptr to start of line */
TOKENPTR  *end  ;                 /* ptr to end of line   */
{
  TOKENPTR tokenptr = *start;
  TOKENPTR prevtoken = NULL;
  
  while (prevtoken != *end) {
   printf("type:%d size:%d font:%d lr:%d width:%d horiz:%d vert:%d\nbegin?:%d prev space width:%d prev space spread:%d\nend?:%d\nnext space width:%d next space spread:%d\nfillers:%d filler_wid:%d letter stretch:%d conn stretch:%d\nfiller stretch:%d\nchar:%c%c%c stretchable?:%d\nstretchability:%c connectivity:%c already_stretched_amount:%d\nalready_stretched_A_conn__amt:%d already_stretched_B_conn__amt:%d",
       tokenptr->token_type,
       tokenptr->point_size,
       tokenptr->font,
       tokenptr->lr,
       tokenptr->width,
       tokenptr->horizontal_pos,tokenptr->vertical_pos,
       tokenptr->begining,
       tokenptr->prev_space_width,
       tokenptr->prev_space_spread,
       tokenptr->ending,
       tokenptr->next_space_width,
       tokenptr->next_space_spread,
       tokenptr->fillers_num,
       tokenptr->filler_width,
       tokenptr->letter_stretch,
       tokenptr->connection_stretch,
       tokenptr->filler_stretch,
       tokenptr->char1,tokenptr->char2,tokenptr->char3,
       tokenptr->stretchable_letter,
       tokenptr->internal_stretchability,
       tokenptr->connectivity,
       tokenptr->already_stretched_amount,
       tokenptr->already_stretched_A_connection_amount,
       tokenptr->already_stretched_B_connection_amount
    );
    if (tokenptr->next == NULL)
       printf(" next:NULL\n\n");
    else
       printf(" next:non-NULL\n\n");
    prevtoken = tokenptr;
    tokenptr = tokenptr->next;
  }
  printf("\n");
}

/* used for debugging ... Dan */
print_token(tokenptr)
TOKENPTR  tokenptr;                 /* ptr to token */
{
   printf("type:%d size:%d font:%d lr:%d width:%d horiz:%d vert:%d\nbegin?:%d prev space width:%d prev space spread:%d\nend?:%d\nnext space width:%d next space spread:%d\nfillers:%d filler_wid:%d letter stretch:%d conn stretch:%d\nfiller stretch:%d\nchar:%c%c%c stretchable?:%d\nstretchability:%c connectivity:%c already_stretched_amount:%d\nalready_stretched_A_conn__amt:%d already_stretched_B_conn__amt:%d",
      tokenptr->token_type,
      tokenptr->point_size,
      tokenptr->font,
      tokenptr->lr,
      tokenptr->width,
      tokenptr->horizontal_pos,tokenptr->vertical_pos,
      tokenptr->begining,
      tokenptr->prev_space_width,
      tokenptr->prev_space_spread,
      tokenptr->ending,
      tokenptr->next_space_width,
      tokenptr->next_space_spread,
      tokenptr->fillers_num,
      tokenptr->filler_width,
      tokenptr->letter_stretch,
      tokenptr->connection_stretch,
      tokenptr->filler_stretch,
      tokenptr->char1,tokenptr->char2,tokenptr->char3,
      tokenptr->stretchable_letter,
      tokenptr->internal_stretchability,
      tokenptr->connectivity,
      tokenptr->already_stretched_amount,
      tokenptr->already_stretched_A_connection_amount,
      tokenptr->already_stretched_B_connection_amount
   );
   if (tokenptr->next == NULL)
      printf(" next:NULL\n\n");
   else
      printf(" next:non-NULL\n\n");
}
