/* (c) Copyright 1985 Cary Buchman and Daniel M. Berry */
/* (c) Copyright 1987 Berry Computer Scientists, Ltd. */
/*********************************************************
**********************************************************
**                                                      **
** this file contains a number of routines that work on **
** internal tokens and token lines                      **
**                                                      **
**********************************************************
*********************************************************/

#include  <math.h>
#include  <stdio.h>
#include  "TOKEN.h"
#include  "TABLE.h"
#include  "macros.h"
#include  "lex.h"



extern int  out_font;                  /* current output font */
extern int  out_size;                  /* current output size */
extern int  out_horizontal;            /* current output horizontal position */
extern int  out_vertical;              /* current output vertical position */
extern char out_font_name[];           /* current output font name */
extern TABLENTRY out_fontable[];       /* currnet output font table */
extern int  in_font;                   /* current input font */
extern int  in_size;                   /* current input size */
extern int  in_horizontal;             /* current input horizontal position */
extern int  in_vertical;               /* current input vertical position */
extern char in_font_name[];            /* current input font name */
extern TABLENTRY in_fontable[];        /* currnet input font table */
extern bool in_lr;                     /* current input font direction */
extern bool direction_table[];         /* direction of font i */
extern bool must_recover_at_word_end;  /* must go back to Vpos of base line */
extern bool in_zero_widths;            /* still processing zero width chars */
extern TOKENPTR beginnings_of_nested_numbers[];
                                       /* stack of pointers to beginnings of  */
                                       /* nested numbers */
extern TOKENPTR predecessors_of_nested_numbers[];
                                       /* stack of pointers to characters */
                                       /* before beginnings of nested numbers */
extern int in_nested_number;           /* top of stack index */
extern bool need_to_push_beginning_of_nested_number;
                                       /* to cause a pointer to the next token*/
                                       /* to be pushed into the beginnings..  */
                                       /* stack                               */
extern double word_length;             /* length of current word              */
extern double word_beginning_x;        /* x coordinate of beginning of        */
                                       /* current word                        */
extern double word_middle_x;           /* x coordinate of middle of           */
                                       /* current word                        */
extern double slant_angle;             /* slant angle of the font of the      */
                                       /* current word                        */
extern int word_beginning_out_vertical;/* output vertical position of begining*/
                                       /* of word */
       double get_slant_angle();       /* returns slant angle of font */

int i;

/****************************************************************
*****************************************************************
**                                                             **
** this routine allocates and initializes a new internal token **
**                                                             **
*****************************************************************
****************************************************************/

TOKENPTR
new_token (toktyp,ptsize,font,horizontal,vertical,fontdir,beg_wrd,end_wrd,char1,
           char2,char3,f_name,width,stretchable_letter,
	   internal_stretchability,connectivity,already_stretched_amount,
	   already_stretched_A_connection_amount,
	   already_stretched_B_connection_amount)

  int          toktyp;         /* token type                 */
  int          ptsize;         /* point size                 */
  int          font;           /* font number                */
  int          horizontal;     /* horizontal position        */
  int          vertical;       /* vertical position          */
  bool         fontdir;        /* font direction             */
  bool         beg_wrd;        /* begining of word indicator */
  bool         end_wrd;        /* end of word indicator      */
  char         char1;          /* 1st char of (abs) character token*/
  char         char2;          /* 2nd char of (abs) character token*/
  char         char3;          /* 3nd char of abs character token*/
  char         *f_name;        /* font name                  */
  int          width;          /* character width            */

  bool         stretchable_letter;        /* if it's one of those       */
  char         internal_stretchability;   /* N or S                     */
  char         connectivity;              /* U, N, P, A, or B
  int          already_stretched_amount;  /* non-zero if character is
                                              manually stretched        */
  int          already_stretched_A_connection_amount;
					  /* non-zero if character's after
					     connection is manually stretched */
  int          already_stretched_B_connection_amount;
					  /* non-zero if character's before
					     connection is manually stretched */
{

  char  *calloc();
  TOKENPTR tokenptr = (TOKENTYPE *) calloc(1,sizeof(TOKENTYPE));

  if (tokenptr != NULL)
  {
    tokenptr->token_type               = toktyp;
    tokenptr->point_size               = ptsize;
    tokenptr->font                     = font;
    strcpy(tokenptr->font_name,f_name);
    tokenptr->lr                       = fontdir;
    tokenptr->width                    = width;
    tokenptr->vertical_pos             = vertical;
    tokenptr->horizontal_pos           = horizontal;
    tokenptr->begining                 = beg_wrd;
    tokenptr->ending                   = end_wrd;
    tokenptr->fillers_num              = NOFILLERS;
    tokenptr->filler_width             = 0;
    tokenptr->letter_stretch           = 0;
    tokenptr->connection_stretch       = 0;
    tokenptr->filler_stretch           = 0;
    tokenptr->char1                    = char1;
    tokenptr->char2                    = char2;
    tokenptr->char3                    = char3;

    tokenptr->stretchable_letter       = stretchable_letter;
    tokenptr->internal_stretchability  = internal_stretchability;
    tokenptr->connectivity             = connectivity;
    tokenptr->already_stretched_amount = already_stretched_amount;
    tokenptr->already_stretched_A_connection_amount =
	already_stretched_A_connection_amount;
    tokenptr->already_stretched_B_connection_amount =
	already_stretched_B_connection_amount;

    tokenptr->prev_space_width = 0;
    tokenptr->prev_space_spread = 0;
    tokenptr->next_space_width = 0;
    tokenptr->next_space_spread = 0;

    tokenptr->next          = NULL;

  }
  else
  {
    out_of_memory();
  }
 return (tokenptr);
}


/************************************************************************/
/************************************************************************/


/*
    this routine adds a token to the end of a line 
                                                     */

add_token (tokenptr,start,end)

  TOKENPTR     tokenptr;           /* token to be added        */
  TOKENPTR     *start;             /* ptr to start of line ptr */
  TOKENPTR     *end;               /* ptr to end of line ptr   */

{
  int i;
  if (*start == NULL) {
    *start = tokenptr;
/*===================================================*/
    if (in_nested_number >= 0) {
        for (i=0; i <= in_nested_number; i++) {
	    predecessors_of_nested_numbers[i] = NULL;
	    beginnings_of_nested_numbers[i] = tokenptr;
        }
    }
/*===================================================*/
  } else {
    (*end)->next = tokenptr;
	tokenptr->next = NULL;        /* better to make sure it's NULL */
/*===================================================*/
    if (need_to_push_beginning_of_nested_number) {
	beginnings_of_nested_numbers[in_nested_number] = tokenptr;
	need_to_push_beginning_of_nested_number = FALSE;
    }
/*===================================================*/
  }
  *end = tokenptr;
  return;
}


/************************************************************************/
/************************************************************************/

/*
    this routine pushes a token onto the front of a line 
                                                           */
push_token (tokenptr,start,end)

  TOKENPTR     tokenptr;            /* token to be push         */
  TOKENPTR     *start;              /* ptr to start of line ptr */
  TOKENPTR     *end;                /* ptr to end of line ptr   */

{

  if (*end == NULL) {
				/* reset the prev and next pointers ...
				   it's not sure that we call it ONLY
				   after new_token.  Johny  */
	(tokenptr)->next = NULL;
	*end = tokenptr;
  }
  else {
	tokenptr->next = (*start);
  }
  *start = tokenptr;
  return;
}

/*******************************************
********************************************
**                                        **
** this routine outputs an internal token ** 
**                                        **
********************************************
*******************************************/

put_token (tokenptr,start_of_line)

  TOKENPTR tokenptr;           /* token to output */
  bool     start_of_line;      /* indicates first character in line */

{
  int offset;
/*int at_this_time; */

  if (start_of_line)
  {
    printf("H%d\n",tokenptr->horizontal_pos);
  }
 
  if (tokenptr->point_size != out_size)
  {
    printf ("s%d\n",tokenptr->point_size);
    out_size = tokenptr->point_size;
  }
  if (strcmp(tokenptr->font_name,out_font_name))
  {
    if (strcmp(tokenptr->font_name,out_fontable[tokenptr->font].name))
    {
      printf("x font %d %s\n",tokenptr->font,tokenptr->font_name);
      new_font(tokenptr->font,tokenptr->font_name,tokenptr->lr,
               out_fontable);
    }
    printf ("f%d\n",tokenptr->font);
    out_font = tokenptr->font;
    strcpy(out_font_name,tokenptr->font_name);
  } 

  if (tokenptr->vertical_pos != out_vertical)
  {
    printf ("V%d\n",tokenptr->vertical_pos);
    out_vertical = tokenptr->vertical_pos;
  }

  slant_angle=get_slant_angle(out_font);
  if (slant_angle != 0.0) {
      if (tokenptr->begining){
	  word_length = (double)(Word_length(tokenptr));
	  offset=(int)((tan(slant_angle*M_PI/180.0)* word_length/2.0)+.5);
	  printf ("V%d\n",out_vertical+offset);
	  if (out_vertical != tokenptr->vertical_pos) fprintf(stderr,
	      "\nCurrent out_vertical != current token's vertical_pos\n");
	  must_recover_at_word_end=TRUE;
	  word_beginning_x = (double)(tokenptr->horizontal_pos);
	  word_middle_x = (word_length/2.0) + word_beginning_x;
	  word_beginning_out_vertical = out_vertical;
      }
  }

  if (!start_of_line)
  {
    if (tokenptr->horizontal_pos < out_horizontal)
    {
      printf("H%d\n",tokenptr->horizontal_pos);
      if (must_recover_at_word_end && slant_angle != 0.0 &&
	      tokenptr->width == 0) {
	  in_zero_widths = TRUE;
	  offset = (int)((tan(slant_angle*M_PI/180.0)* (double)(word_middle_x
	            - tokenptr->horizontal_pos))+.5);
          printf ("V%d\n",out_vertical+offset);
	  if (out_vertical != tokenptr->vertical_pos) fprintf(stderr,
	      "\nCurrent out_vertical != current token's vertical_pos\n");
	  out_vertical = word_beginning_out_vertical;
      }
    }
    else if (tokenptr->horizontal_pos > out_horizontal) {
	if (in_zero_widths && tokenptr->width == 0 && slant_angle != 0.0) {
            printf("H%d\n",tokenptr->horizontal_pos);
	    offset = (int)((tan(slant_angle*M_PI/180.0)* (double)(word_middle_x
	              - tokenptr->horizontal_pos))+.5);
            printf ("V%d\n",out_vertical+offset);
	    if (out_vertical != tokenptr->vertical_pos) fprintf(stderr,
	        "\nCurrent out_vertical != current token's vertical_pos\n");
	    out_vertical = word_beginning_out_vertical;
	} else {
          int j =(tokenptr->horizontal_pos - out_horizontal);
          if ((j < 100 && j > 9 && tokenptr->token_type == c_token) &&
              (tokenptr->letter_stretch == 0) &&
              (tokenptr->already_stretched_amount == 0)){
                  printf( "%d%c", j, tokenptr->char1);
                  goto compress;
              }
          printf("h%d",(tokenptr->horizontal_pos - out_horizontal));
	}
    }
  }

  if (tokenptr->letter_stretch > 0 && tokenptr->already_stretched_amount == 0) {
	printf("x X stretch %d\n",tokenptr->letter_stretch);
  }

  if (tokenptr->already_stretched_amount > 0 && tokenptr->letter_stretch == 0) {
	printf("x X stretch %d\n",tokenptr->already_stretched_amount);
  }

  if (tokenptr->letter_stretch > 0 && tokenptr->already_stretched_amount > 0) {
	printf("x X stretch %d\n",tokenptr->letter_stretch +
				tokenptr->already_stretched_amount);
  }

  if (/*tokenptr->connection_stretch == 0 && */
	  tokenptr->already_stretched_B_connection_amount > 0){
      printf("x X BCstretch %d\n",
             tokenptr->already_stretched_B_connection_amount);
  }
  if (tokenptr->already_stretched_A_connection_amount > 0){
      printf("x X ACstretch %d\n",
             tokenptr->already_stretched_A_connection_amount);
  }

  switch (tokenptr->token_type) {
    case c_token: printf("c%c\n",tokenptr->char1); break;
    case C_token: printf("C%c%c\n",tokenptr->char1,tokenptr->char2); break;
    case N_token: printf("N%c%c%c\n",tokenptr->char1,tokenptr->char2,
						tokenptr->char3); break;
  }
compress: /* thanks to Mulli Bahr hnncx compressed to nnx */
  out_horizontal = tokenptr->horizontal_pos;

  /* added to insert the fillers ... Johny */
  if (tokenptr->fillers_num > 0) {
	int i;
	printf("h%d",tokenptr->width);
	printf("Chy\n");
	for (i=0; i<(tokenptr->fillers_num)-1; i++)
		printf("h%dChy\n",tokenptr->filler_width);
	printf("h%d",tokenptr->filler_width);
	out_horizontal += tokenptr->fillers_num * tokenptr->filler_width +
					  tokenptr->width;
  }

  if (tokenptr->filler_stretch > 0) {
          printf("h%d", tokenptr->width);
          printf("x X stretch %d\n", tokenptr->filler_stretch);
          printf("Chy\n");
          printf("h%d", tokenptr->filler_stretch);
          out_horizontal+=tokenptr->filler_stretch+tokenptr->width;
  }

  if ((tokenptr->ending == END) && (tokenptr->next !=NULL)) {
    printf("w");
    if (must_recover_at_word_end){
	printf("V%d\n",out_vertical);
	must_recover_at_word_end = FALSE;
	in_zero_widths = FALSE;
	word_length = 0.0;
	word_beginning_x = 0.0;
	word_middle_x = 0.0;
	word_beginning_out_vertical = 0;
	slant_angle = 0.0;
    }
  }

  return;
}


/*************************************************************
**************************************************************
**                                                          **
** this routine outputs a new page token and its associated **
** font and point size tokens and resets its font and size  **
** to null and 0 so as to force output of f and s commands  **
**                                                          **
**************************************************************
*************************************************************/

put_page_token(page_token)

  char  *page_token;          /* lex page token value */

{

  extern char  yytext[];     /* lex token value */
  
  printf("V%d\n",in_vertical); 
				/* ditroff guarantees and thus drivers
				require that before page_token, a V0 token! 
				Mulli Bahr, Oct, 86*/
  out_vertical = in_vertical;
  printf("%s\n",page_token);

  out_size = 0;                            /* reset out_size to zero ... */
  out_font_name[0] = '\0';                 /* reset out_font_name to empty
         string  so that the next token print will be forced to dump out s and f
         commands. On some devices the current size and page are not remembered
         across page boundaries, especially if there's a bunch of "x font XX"s.
         Besides the ditroff output always seems to have the s and f issued just
         before the text on each page */

  return;
}


/***********************************************************************/
/***********************************************************************/


/*
    this routine deallocates a line of tokens
                                                */

free_line(start,end)

  TOKENPTR  *start;           /* ptr to start of line */
  TOKENPTR  *end;             /* ptr to end of line   */

{
  TOKENPTR  tokenptr;         /* next element in line */

  while( (tokenptr = *start) != NULL)
  {
    *start = tokenptr->next;
    cfree(tokenptr);
  }
  *end = NULL;
  return;
}

/************************************************************************/ 
/************************************************************************/ 

